#include "puissance4.h"
#include <stdint.h>
#include <stdio.h>

/******************************************************************
 * function     : Create_grid2D
 * arguments    : taille de grille hauteur x largeur
 *
 * return value : cell** -> la grille mise à jour et allouée
 * 
 * Description : initialise les valeurs de la grille et alloue la 
 * mémoire dynamiquement
 ******************************************************************/
struct cell **Create_grid2D(int height, int width){
	struct cell** cells = malloc(width*sizeof(struct cell*));
	for(int k =0; k<width; k++){
		cells[k] = malloc(height*sizeof(struct cell));
	}
	//initialise les positions et les valeurs de base
	for(int i = 0; i<width; i++){
		for(int j=0; j<height; j++){
			cells[i][j].symbol = EMPTY;
			cells[i][j].i_pos = i;
			cells[i][j].j_pos = j;
		}
	}
	return cells;
}


/******************************************************************
 * function     : init_puissance4
 * arguments    : la structure du jeu struct game
 * 				  hauteur du plateau virtuel
 * 				  largeur du plateau virtuel
 *
 * return value : 0 -> si tout a fonctionné
 * 				 -1 -> si erreur
 * 
 * Description : prends le jeu alloué en paramètre, puis l'initialiser
 * en tenant compte de la taille du plateau souhaité
 ******************************************************************/
int init_puissance4(struct game *game, int height, int width){
	if(game == NULL){
		return -1;
	}
	if(height > 3 || width > 3){
		game->players[1].check_win = 0;
		game->players[1].score = 0;
		game->players[1].symbol = CIRCLE;

		game->players[0].check_win = 0;
		game->players[0].score = 0;
		game->players[0].symbol = CROSS;

		game->curr_player = 0;
		game->height = height;
		game->width = width;
		game->gamePlayed = height*width;
		game->cells = Create_grid2D(height, width);
		return 0;
	}else{
		return -1;
	}
}

/******************************************************************
 * function     : print_cells
 * arguments    : le tableau 2D type cell**, hauteur, largeur
 *
 * return value : -
 * Description : affiche simplement le contenu du tableau 2D cell**
 ******************************************************************/
void print_cells(struct cell **cell, int height, int width){
	for(int i=0; i<width; i++){
		for(int j=0; j<height; j++){
			printf("%d, ", cell[i][j].symbol);
		}
		printf("\n");
	}
}

/******************************************************************
 * function     : print_gameCells
 * arguments    : structure du jeu "struct game"
 *
 * return value : -
 * Description : affiche le contenu du tableau 2D cell** contenu dans
 * la structure game
 ******************************************************************/
void print_gameCells(struct game game){
	print_cells(game.cells, game.height, game.width);
}

/******************************************************************
 * function     : cell_destroy
 * arguments    : tableau 2D struct cell**
 * 				  largeur du tableau
 *
 * return value : -
 * 
 * Description : détruit et libère la mémoire allouée sur le tableau
 * 2D du tableau alloué avec Create_grid2D
 ******************************************************************/
void cell_destroy(struct cell **cells, int width){
	for(int i=0; i<width; i++){
		free(cells[i]);
	}
	free(cells);
}

/******************************************************************
 * function     : kill_game
 * arguments    : struct game** (structure du jeu)
 *
 * return value : 0 -> tout fonctionne
 * 				 -1 -> erreur
 * 
 * Description : détruit et libère la mémoire allouée de la structure
 * du jeu si non NULL.
 ******************************************************************/
int kill_game(struct game **game){
	if(*game==NULL){
		return -1;
	}
	cell_destroy((*game)->cells, (*game)->width);
	free(*game);
	*game=NULL;
	return EXIT_SUCCESS;
}

/******************************************************************
 * function     : put_free_cell
 * arguments    : struct game* (structure du jeu)
 * 				  int j_p -> colonne sélectionnée par le joueur
 * 				  symbol_t symbol -> symbole du joueur
 *
 * return value : [0-x] -> valeur de la ligne du tableau où placer le jeton
 * 				  -1	-> il n'y a pas de case disponible dans la colonne selectionée
 * 
 * Description : place le jeton du joueur correspondant au symbol_t
 * dans la case j_p, puis le place dans la dernière case vide si il
 * y en a une.
 ******************************************************************/
int put_free_cell(struct game *game, int j_p, symbol_t symbol){
	int i=0;
	if((i = is_cell_free(game, j_p, 0, game->players[game->curr_player].symbol)) != -1){
		game->cells[j_p][i].symbol = symbol;
		game->gamePlayed -= 1;
	}
	return i;
}

/******************************************************************
 * function     : is_cell_free
 * arguments    : struct game* (structure du jeu)
 * 				  int j_p -> colonne sélectionnée par le joueur
 * 				  int i -> doit être = à 0 au premier appel ! pour la recursion
 * 				  symbol_t symbol -> symbole du joueur
 *
 * return value : [0-x] -> valeur de la ligne du tableau où placer le jeton
 * 				  -1	-> il n'y a pas de case disponible dans la colonne selectionée
 * 
 * Description : trouve dans la colonne donnée la ligne disponible
 * pour placer le jeton.
 ******************************************************************/
int is_cell_free(struct game *game, int j_p, int i, symbol_t symbol){
	if(game->cells[j_p][i].symbol == EMPTY){
		//retourne la ligne disponible
		return i;
	}else{
		if(i<(game->height-1)){
			return is_cell_free(game, j_p, i+1, symbol);
		}else{
			return -1;
		}
	}
}

/******************************************************************
 * function     : Launch_puissance4
 * arguments    : struct game* (structure du jeu)
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode 2players (deux joueurs)
 ******************************************************************/
int Launch_puissance4(struct game *game){
	symbol_t winner=EMPTY;
	int chiffre=0;
	if(game == NULL){
		return -1;
	}
	while(winner==EMPTY){
		print_game(*game);
		printf("Column number? (starts at 1):\n");
		do{
			scanf("%d",&chiffre);
			chiffre-=1; //gestion des chiffres de 1 à n et non de 0 à n-1
			if(chiffre >= game->width || chiffre < 0){
				printf("HEP HEPHEP !! donne un chiffre dans la plage de chiffres indiquée !\n");
			}
		}while((chiffre >= game->width) || (chiffre < 0) || is_cell_free(game, chiffre, 0, game->players[game->curr_player].symbol)<0);
		int i = 0;
		//verifier que put free cell fonctionne
		if((i = put_free_cell(game, chiffre, game->players[game->curr_player].symbol))<0){
			if(Is_Grid_full(*game)){
				//print egalite mais y a probablement eu un bug
				printf("y a un egalite mais probablement un bug\n");
				//fin du jeu
				return EQUAL;
			}else{
				printf("la y a un bug, faut check ca \n");
				print_game(*game);
				//fin du jeu
				return -1; //retourne une erreur
			}
		}
		//check le winner
		winner = Find_winner(game, game->cells, game->cells[chiffre][i]);
		if( winner != EMPTY){
			if(winner==CROSS){
				//reset du jeu + mise a jour des scores (scores optionels)
				print_game(*game);
				printf("Player one won!\n");
				return CROSS;
			}
			if(winner==CIRCLE){
				//reset du jeu + mise a jour des scores (scores optionels)
				print_game(*game);
				printf("Player two won!\n");
				return CIRCLE;
			}
		}else{
			//sois le jeu est plein, sois y a juste pas encore de winner
			if(Is_Grid_full(*game)){
				//la grille est pleine
				print_game(*game);
				printf("It is a draw.\n");
				return EQUAL;
			}
			//y a juste pas encore de winner ducoup...
			game->curr_player ^= 1; //switch de joueur
		}
	}
	return winner;
}

/******************************************************************
 * function     : Launch_puissance4_randBot
 * arguments    : struct game* (structure du jeu)
 * 				  int seed -> graine de generation pseudo-aléatoire
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode joueur vs bot,
 * mais le bot a un algorithme pseudo aléatoire.
 ******************************************************************/
int Launch_puissance4_randBot(struct game *game, int seed){
	if(game == NULL){
		return -1;
	}
	symbol_t winner=EMPTY;
	int chiffre=0;
	srand(seed);

	while(winner==EMPTY){
		print_game(*game);
		if(game->curr_player == 1){
			do{
				chiffre = rand()%game->width;
			}while((is_cell_free(game, chiffre, 0, game->players[game->curr_player].symbol))<0);
		}else{
			printf("Column number? (starts at 1):\n");
			do{
				scanf("%d",&chiffre);
				chiffre-=1; //gestion des chiffres de 1 à n et non de 0 à n-1
				if(chiffre >= game->width || chiffre < 0){
					printf("HEP HEPHEP !! donne un chiffre dans la plage de chiffres indiquée !\n");
				}
			}while((chiffre >= game->width) || (chiffre < 0) || is_cell_free(game, chiffre, 0, game->players[game->curr_player].symbol)<0);
		}
		int i = 0;
		//verifier que put free cell fonctionne
		if((i = put_free_cell(game, chiffre, game->players[game->curr_player].symbol))<0){
			if(Is_Grid_full(*game)){
				printf("y a un egalite mais probablement un bug\n");
				return EQUAL;
			}else{
				printf("ok la y a un bug de zinzin check ca \n");
				print_game(*game);
				//fin du jeu
				return -1; //erreur
			}
		}
		//check de la win
		winner = Find_winner(game, game->cells, game->cells[chiffre][i]);
		if( winner != EMPTY){
			if(winner==CROSS){
				print_game(*game);
				printf("Player one won!\n");
				//fin du jeu
				return CROSS;
			}
			if(winner==CIRCLE){
				print_game(*game);
				printf("Computer won!\n");
				//fin du jeu
				return CIRCLE;
			}
		}else{
			//sois le jeu est plein, sois y a juste pas encore de winner
			if(Is_Grid_full(*game)){
				print_game(*game);
				printf("It is a draw.\n");
				//fin du jeu
				return EQUAL;
			}
			//y a juste pas encore de winner ducoup...
			game->curr_player ^= 1; //switch de joueur
		}
	}
	return winner;
}

/******************************************************************
 * function     : Launch_puissance4_smartBot
 * arguments    : struct game* (structure du jeu)
 * 				  int seed -> graine de generation pseudo-aléatoire
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode joueur vs smart-bot,
 * le bot peux trouver un coup gagnant si il peux en jouer un,
 * bloquer un coup du joueur adversaire si il peux
 * sinon, joue en mode pseudo-aléatoire.
 ******************************************************************/
int Launch_puissance4_smartBot(struct game *game, int seed){
	if(game == NULL){
		return -1;
	}
	symbol_t winner=EMPTY;
	int chiffre=0;
	srand(seed);

	while(winner==EMPTY){
		print_game(*game);
		if(game->curr_player == 1){
			chiffre = SmartBot(game);
		}else{
			printf("Column number? (starts at 1):\n");
			do{
				scanf("%d",&chiffre);
				chiffre-=1; //gestion des chiffres de 1 à n et non de 0 à n-1
				if(chiffre >= game->width || chiffre < 0){
					printf("HEP HEPHEP !! donne un chiffre dans la plage de chiffres indiquée !\n");
				}
			}while((chiffre >= game->width) || (chiffre < 0) || is_cell_free(game, chiffre, 0, game->players[game->curr_player].symbol)<0);
		}
		int i = 0;
		//verifier que put free cell fonctionne
		if((i = put_free_cell(game, chiffre, game->players[game->curr_player].symbol))<0){
			if(Is_Grid_full(*game)){
				printf("y a un egalite mais probablement un bug\n");
				//fin du jeu
				return EQUAL;
			}else{
				printf("il y a un bug, faut check ca \n");
				print_game(*game);
				//fin du jeu
				return -1; //erreur
			}
		}
		//check de la win
		winner = Find_winner(game, game->cells, game->cells[chiffre][i]);
		if( winner != EMPTY){
			if(winner==CROSS){
				print_game(*game);
				printf("Player one won!\n");
				return CROSS;
			}
			if(winner==CIRCLE){
				print_game(*game);
				printf("Computer won!\n");
				return CIRCLE;
			}
		}else{
			//sois le jeu est plein, sois y a juste pas encore de winner
			if(Is_Grid_full(*game)){
				print_game(*game);
				printf("It is a draw.\n");
				return EQUAL;
			}
			//y a juste pas encore de winner ducoup...
			game->curr_player ^= 1; //switch de joueur
		}
	}
	return winner;
}

/******************************************************************
 * function     : SmartBot
 * arguments    : struct game* (structure du jeu)
 *
 * return value : retourne le prochain coup joué par le bot
 * 
 * Description : algorithme smartBot utilisé dans le mode de jeu 2
 * le bot peux trouver un coup gagnant si il peux en jouer un,
 * bloquer un coup du joueur adversaire si il peux
 * sinon, joue en mode pseudo-aléatoire.
 ******************************************************************/
int SmartBot(struct game *game){
	int pos=0;
	int chfr1 = -1, chfr2 = -1;
	int game_played = game->gamePlayed;
	for(int chfr=0; chfr<game->width; chfr++){
		//verifie que le coup est jouable
		if((pos = put_free_cell(game, chfr, game->players[game->curr_player].symbol))>=0){
			struct cell *tempCell = &game->cells[chfr][pos];
			tempCell->symbol = game->players[game->curr_player].symbol;
			if (Find_winner(game, game->cells, game->cells[chfr][pos]) == game->players[game->curr_player].symbol){
				 //1er cas : a trouvé le bon coup à jouer pour gagner
				tempCell->symbol = EMPTY;
				chfr1 = chfr;
			}
			//tente de chercher si l'autre joueur a un coup gagnant
			tempCell->symbol = game->players[game->curr_player^1].symbol;
			if (Find_winner(game, game->cells, game->cells[chfr][pos]) == game->players[game->curr_player^1].symbol){
				//2e cas : joue le coup gagnant de l'autre joueur pour le bloquer
				tempCell->symbol = EMPTY;
				chfr2 = chfr;
			}
			tempCell->symbol = EMPTY;
		}
	}
	game->gamePlayed = game_played;
	if(chfr1 != -1){
		return chfr1;
	}else if(chfr2 != -1){
		return chfr2;
	}else{
		//3e cas : joue un coup au hasard
		int randomchfr=0;
		do{
			randomchfr = rand()%game->width;
		}while((is_cell_free(game, randomchfr, 0, game->players[game->curr_player].symbol))<0);
		return randomchfr;
	}
}

/******************************************************************
 * function     : Is_Grid_full
 * arguments    : struct game (structure du jeu)
 *
 * return value : retourne si la partie est egalité ou non
 * 				  (si la grille est pleine ou non)
 * 
 * Description : Lorsque le jeu est lancé, cette fonction est utilisée
 * pour savoir si la grille de jeu est pleine (ou si c'est un égalité)
 ******************************************************************/
bool Is_Grid_full(struct game game){
	return game.gamePlayed == 0;
}

/******************************************************************
 * function     : Find_winner
 * arguments    : struct game (structure du jeu)
 * 				  struct cell** : grille a tester
 * 				  struct cell : cellule jouée,
 *
 * return value : symbole -> gagnant ou EMPTY
 * 
 * Description : 
 * Cherche si il y a un gagnant selon le dernier coup
 * joué.
 * !!! Fonctionne avec CheckWin_in_a_direction !!!
 ******************************************************************/
symbol_t Find_winner(struct game *game, struct cell **grid, struct cell cellPlayed){
	//int grid_len = game.height*game.width;
	int k[4][2]={
			{0,1},
			{1,0},
			{1,-1},
			{1,1}
	};
	int k2[4][2]={
			{0,-1},
			{-1,0},
			{-1,1},
			{-1,-1}
	};
	symbol_t result;
	for(int i=0; i<4; i++){
		//à savoir, il est impossible que le cellPlayed soit empty !!
		result = CheckWin_in_a_direction(game, k[i], grid, grid[cellPlayed.i_pos][cellPlayed.j_pos]);
		if(result != EMPTY){return result;}
		result = CheckWin_in_a_direction(game, k2[i], grid, grid[cellPlayed.i_pos][cellPlayed.j_pos]);
		game->players[0].check_win = 0;
		game->players[1].check_win = 0;
		if(result != EMPTY){
			return result; //si il trouve un winner, alors il stoppe la fonction :)
		}
	}
	return EMPTY; // pas de winner pour l'instant
}

/******************************************************************
 * function     : CheckWin_in_a_direction
 * arguments    : struct game* (structure du jeu)
 * 				  int dir[2] : direction 2D où chercher une suite
 * 				  struct cell : cellule jouée
 *
 * return value : symbole -> gagnant ou EMPTY si non
 *
 * Description : Va chercher des suites de cases gagnantes
 * recursivement, dans une seule direction selon le dernier jeton joué
 ******************************************************************/
symbol_t CheckWin_in_a_direction(struct game *game, int dir[2], struct cell **grid, struct cell cell){
	int x1 = cell.i_pos+dir[0];
	int x2 = cell.j_pos+dir[1];
	symbol_t result = EMPTY;

	if(x1<game->width && x1>=0 && x2<game->height && x2>=0){
		if(grid[x1][x2].symbol==cell.symbol){
			if(cell.symbol == CIRCLE){
				game->players[0].check_win += 1;
			}else{
				game->players[1].check_win += 1;
			}
			if(game->players[0].check_win>=3){
				return CIRCLE;	//circle win
			}
			else if(game->players[1].check_win>=3){
				return CROSS;   //cross win
			}else{
				result = CheckWin_in_a_direction(game, dir, grid, grid[x1][x2]);
			}
		}
	}
	return result;
}

/******************************************************************
 * function     : print_game
 * arguments    : struct game (structure du jeu)
 *
 * return value : -
 * Description : fonction d'affichage du jeu de la manière suivante :
 * (ex pour un jeu 6x7)
 * ┌─┬─┬─┬─┬─┬─┬─┐
 * │ │ │ │ │ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │ │ │ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │ │X│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │X│O│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │O│X│X│O│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │X│O│X│O│ │ │ │
 * └─┴─┴─┴─┴─┴─┴─┘
 *  1 2 3 4 5 6 7
 *
 ******************************************************************/
void print_game(struct game game){
	//printf("h:%d, w:%d\n", game.height, game.width);
	//affiche ┌─┬─┬─┬─┬─┬─┬─┐
	for(int i=0; i<(game.width*2)+1; i++){
		if(i%2){
			printf("─");
		}else{
			if(i==0){
				printf("┌");
			}
			else if(i == (game.width*2)){
				printf("┐");
			}
			else{
				printf("┬");
			}
		}
	}
	printf("\n");

	//affiche le jeu-> │ │ │X│O│ │ │ │
	for(int i=0; i<game.height; i++){
		printf("│");
		for(int j=0; j<game.width; j++){
			if(game.cells[j][(game.height-i)-1].symbol==EMPTY){
				printf(" ");
			}
			else if(game.cells[j][(game.height-i)-1].symbol==CROSS){
				printf("X");
			}else if(game.cells[j][(game.height-i)-1].symbol==CIRCLE){
				printf("O");
			}else{
				printf("E");
			}
			printf("│");
		}
		printf("\n");
		if(i<game.height-1){
			//Dessine ├─┼─┼─┼─┼─┼─┼─┤ ...
			for(int cnt=0; cnt<((game.width*2)+1); cnt++){
				if(cnt%2){
					printf("─");
				}else{
					if(cnt == 0){
						printf("├");
					}
					else if(cnt == (game.width*2)){
						printf("┤");
					}
					else{
						printf("┼");
					}
				}
			}
			printf("\n");
		}
	}

	//Affiche └─┴─┴─┴─┴─┴─┴─┘
	for(int i=0; i<(game.width*2)+1; i++){
		if(i%2){
			printf("─");
		}else{
			if(i==0){
				printf("└");
			}
			else if(i == (game.width*2)){
				printf("┘");
			}
			else{
				printf("┴");
			}
		}
	}
	printf("\n");

	//affiche les numéros
	int cnt = 0;
	for(int i=0; i<(game.width*2); i++){
		if(i%2){
			printf("%d",++cnt);
		}else{
			printf(" ");
		}
	}
	printf("\n");
}