#ifndef _PUIS4_
#define _PUIS4_

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

typedef enum{
	CIRCLE,	//signifie le symbole du cercle
	CROSS,	//signifie le symbole de la croix
	EMPTY,	//signifie une cellule vide
	EQUAL	//signifie un état d'égalité entre la croix et le cercle
}symbol_t;

struct cell{
    symbol_t symbol;
    int i_pos;
    int j_pos;
};

struct player{
    symbol_t symbol;
    int score;
    int check_win;
};

struct game{
    int height;
    int width;
    int gamePlayed;
    struct cell** cells;
    struct player players[2];
    int curr_player;
};

/******************************************************************
 * function     : Create_grid2D
 * arguments    : taille de grille hauteur x largeur
 *
 * return value : cell** -> la grille mise à jour et allouée
 * 
 * Description : initialise les valeurs de la grille et alloue la 
 * mémoire dynamiquement
 ******************************************************************/
struct cell **Create_grid2D(int height, int width);


/******************************************************************
 * function     : init_puissance4
 * arguments    : la structure du jeu struct game
 * 				  hauteur du plateau virtuel
 * 				  largeur du plateau virtuel
 *
 * return value : 0 -> si tout a fonctionné
 * 				 -1 -> si erreur
 * 
 * Description : prends le jeu alloué en paramètre, puis l'initialiser
 * en tenant compte de la taille du plateau souhaité
 ******************************************************************/
int init_puissance4(struct game *game, int height, int width);

/******************************************************************
 * function     : print_cells
 * arguments    : le tableau 2D type cell**, hauteur, largeur
 *
 * return value : -
 * Description : affiche simplement le contenu du tableau 2D cell**
 ******************************************************************/
void print_cells(struct cell **cell, int height, int width);

/******************************************************************
 * function     : print_gameCells
 * arguments    : structure du jeu "struct game"
 *
 * return value : -
 * Description : affiche le contenu du tableau 2D cell** contenu dans
 * la structure game
 ******************************************************************/
void print_gameCells(struct game game);

/******************************************************************
 * function     : cell_destroy
 * arguments    : tableau 2D struct cell**
 * 				  largeur du tableau
 *
 * return value : -
 * 
 * Description : détruit et libère la mémoire allouée sur le tableau
 * 2D du tableau alloué avec Create_grid2D
 ******************************************************************/
void cell_destroy(struct cell **cells, int width);

/******************************************************************
 * function     : kill_game
 * arguments    : struct game** (structure du jeu)
 *
 * return value : 0 -> tout fonctionne
 * 				 -1 -> erreur
 * 
 * Description : détruit et libère la mémoire allouée de la structure
 * du jeu si non NULL.
 ******************************************************************/
int kill_game(struct game **game);

/******************************************************************
 * function     : put_free_cell
 * arguments    : struct game* (structure du jeu)
 * 				  int j_p -> colonne sélectionnée par le joueur
 * 				  symbol_t symbol -> symbole du joueur
 *
 * return value : [0-x] -> valeur de la ligne du tableau où placer le jeton
 * 				  -1	-> il n'y a pas de case disponible dans la colonne selectionée
 * 
 * Description : place le jeton du joueur correspondant au symbol_t
 * dans la case j_p, puis le place dans la dernière case vide si il
 * y en a une.
 ******************************************************************/
int put_free_cell(struct game *game, int j_p, symbol_t symbol);

/******************************************************************
 * function     : is_cell_free
 * arguments    : struct game* (structure du jeu)
 * 				  int j_p -> colonne sélectionnée par le joueur
 * 				  int i -> doit être = à 0 au premier appel ! pour la recursion
 * 				  symbol_t symbol -> symbole du joueur
 *
 * return value : [0-x] -> valeur de la ligne du tableau où placer le jeton
 * 				  -1	-> il n'y a pas de case disponible dans la colonne selectionée
 * 
 * Description : trouve dans la colonne donnée la ligne disponible
 * pour placer le jeton.
 ******************************************************************/
int is_cell_free(struct game *game, int j_p, int i, symbol_t symbol);

/******************************************************************
 * function     : Launch_puissance4
 * arguments    : struct game* (structure du jeu)
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode 2players (deux joueurs)
 ******************************************************************/
int Launch_puissance4(struct game *game);

/******************************************************************
 * function     : Launch_puissance4_randBot
 * arguments    : struct game* (structure du jeu)
 * 				  int seed -> graine de generation pseudo-aléatoire
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode joueur vs bot,
 * mais le bot a un algorithme pseudo aléatoire.
 ******************************************************************/
int Launch_puissance4_randBot(struct game *game, int seed);

/******************************************************************
 * function     : Launch_puissance4_smartBot
 * arguments    : struct game* (structure du jeu)
 * 				  int seed -> graine de generation pseudo-aléatoire
 *
 * return value : retourne le resultat du gagnant de la partie
 * 				  ou -1 si il y a une erreur
 * 
 * Description : Jeu du puissance4 en mode joueur vs smart-bot,
 * le bot peux trouver un coup gagnant si il peux en jouer un,
 * bloquer un coup du joueur adversaire si il peux
 * sinon, joue en mode pseudo-aléatoire.
 ******************************************************************/
int Launch_puissance4_smartBot(struct game *game, int seed);

/******************************************************************
 * function     : SmartBot
 * arguments    : struct game* (structure du jeu)
 *
 * return value : retourne le prochain coup joué par le bot
 * 
 * Description : algorithme smartBot utilisé dans le mode de jeu 2
 * le bot peux trouver un coup gagnant si il peux en jouer un,
 * bloquer un coup du joueur adversaire si il peux
 * sinon, joue en mode pseudo-aléatoire.
 ******************************************************************/
int SmartBot(struct game *game);

/******************************************************************
 * function     : Is_Grid_full
 * arguments    : struct game (structure du jeu)
 *
 * return value : retourne si la partie est egalité ou non
 * 				  (si la grille est pleine ou non)
 * 
 * Description : Lorsque le jeu est lancé, cette fonction est utilisée
 * pour savoir si la grille de jeu est pleine (ou si c'est un égalité)
 ******************************************************************/
bool Is_Grid_full(struct game game);

/******************************************************************
 * function     : Find_winner
 * arguments    : struct game (structure du jeu)
 * 				  struct cell** : grille a tester
 * 				  struct cell : cellule jouée,
 *
 * return value : symbole -> gagnant ou EMPTY
 * 
 * Description : 
 * Cherche si il y a un gagnant selon le dernier coup
 * joué.
 * !!! Fonctionne avec CheckWin_in_a_direction !!!
 ******************************************************************/
symbol_t Find_winner(struct game *game, struct cell **grid, struct cell cellPlayed);

/******************************************************************
 * function     : CheckWin_in_a_direction
 * arguments    : struct game* (structure du jeu)
 * 				  int dir[2] : direction 2D où chercher une suite
 * 				  struct cell : cellule jouée
 *
 * return value : symbole -> gagnant ou EMPTY si non
 *
 * Description : Va chercher des suites de cases gagnantes
 * recursivement, dans une seule direction selon le dernier jeton joué
 ******************************************************************/
symbol_t CheckWin_in_a_direction(struct game *game, int dir[2], struct cell **grid, struct cell cell);

/******************************************************************
 * function     : print_game
 * arguments    : struct game (structure du jeu)
 *
 * return value : -
 * Description : fonction d'affichage du jeu de la manière suivante :
 * (ex pour un jeu 6x7)
 * ┌─┬─┬─┬─┬─┬─┬─┐
 * │ │ │ │ │ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │ │ │ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │ │X│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │ │ │X│O│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │O│X│X│O│ │ │ │
 * ├─┼─┼─┼─┼─┼─┼─┤
 * │X│O│X│O│ │ │ │
 * └─┴─┴─┴─┴─┴─┴─┘
 *  1 2 3 4 5 6 7
 *
 ******************************************************************/
void print_game(struct game game);

#endif